#define CURL_STATICLIB
#include <curl/curl.h>

#include <zzip/zzip.h>
#include <zzip/file.h>

#define _RETAIL
#include "GMLuaModule.h"
#include "gm_gmodeworld.h"

#include <string>
#include <list>

#include <eiface.h>
#include <filesystem.h>

ILuaShared *lua_shared;
IFileSystem *filesystem;

CURLM *multi_handle;

int num_curl_handles = 0;

struct LuaPrivateData 
{
	int ref_id;
	int ref_finish;
	int ref_progress;
	HANDLE hFile;

	struct curl_httppost *post;
};

ILuaInterface *g_pLua = NULL;

LUA_FUNCTION(curl_poll)
{
	ILuaInterface *gLua = Lua();

	int num_handles;

	while(CURLM_CALL_MULTI_PERFORM ==
		curl_multi_perform(multi_handle, &num_handles));

	for(;;)
	{
		int msgs;
		CURLMsg *msg;

		msg = curl_multi_info_read(multi_handle, &msgs);

		if(msg == NULL || msg->msg == CURLMSG_NONE)
			break;

		if(msg->msg == CURLMSG_DONE)
		{
			LuaPrivateData *data = NULL;
			curl_easy_getinfo(msg->easy_handle, CURLINFO_PRIVATE, &data);

			gLua->PushReference(data->ref_finish);
			gLua->PushReference(data->ref_id);
			gLua->PushLong(msg->data.result);
			gLua->Call(2, 0);

			CURL *handle = msg->easy_handle; // this is important!

			curl_multi_remove_handle(multi_handle, handle);
			curl_easy_cleanup(handle);

			if(data->post)
				curl_formfree(data->post);

			--num_curl_handles;

			gLua->FreeReference(data->ref_finish);
			gLua->FreeReference(data->ref_progress);
			gLua->FreeReference(data->ref_id);

			CloseHandle(data->hFile);

			delete data;
		}
	}
	return 0;
}

size_t write_callback(char *buffer, size_t size, size_t nitems, HANDLE outstream)
{
	DWORD bytesWritten = 0;
	WriteFile(outstream, buffer, size * nitems, &bytesWritten, NULL);
	return bytesWritten;
}

int progress_callback(void *clientp, double dltotal, double dlnow, double ultotal, double ulnow)
{
	LuaPrivateData *data = (LuaPrivateData *)clientp;

	g_pLua->PushReference(data->ref_progress);
	g_pLua->PushReference(data->ref_id);
	g_pLua->PushDouble(dltotal);
	g_pLua->PushDouble(dlnow);

	g_pLua->Call(3, 0);

	return 0;
}

const char *GetZipFile(ILuaInterface *gLua)
{
	static char file[MAX_PATH];

	gLua->CheckType(1, GLua::TYPE_STRING);
	const char *filename = gLua->GetString(1);

	_snprintf(file, sizeof(file), "%saddons\\%s", modulemanager->GetBaseFolder(), filename);
	int len = strlen(file);

	if(strncmp(&file[len-3], "zip", 3) != 0 || strstr(filename, "..") != 0 || strstr(filename, ":") != 0)
		return NULL;

	return file;
}

const char *GetFirstZipDir(ZZIP_DIR *zip, ZZIP_DIRENT *dirent)
{
	if(zzip_dir_read(zip, dirent))
	{
		return dirent->d_name;
	}

	return NULL;
}

ZZIP_DIR *GetZipDir(ILuaInterface *gLua)
{
	const char *file = GetZipFile(gLua);
	if(!file)
		return NULL;

	ZZIP_DIR *zipdir = zzip_opendir(file);
	if(!zipdir)
		return NULL;

	return zipdir;
}

CURL *GetNewHandle(ILuaInterface *gLua)
{
	gLua->CheckType(2, GLua::TYPE_STRING);
	gLua->CheckType(3, GLua::TYPE_FUNCTION);
	gLua->CheckType(4, GLua::TYPE_FUNCTION);

	const char *file = GetZipFile(gLua);

	if(!file)
		return NULL;

	CURL *handle = curl_easy_init();

	curl_easy_setopt(handle, CURLOPT_URL, gLua->GetString(2));
	curl_easy_setopt(handle, CURLOPT_USERAGENT, "Gmode/11.0");

	int reference_finished = gLua->GetReference(3);
	int reference_progress = gLua->GetReference(4);
	int ref_id = gLua->GetReference(5);

	LuaPrivateData *data = new LuaPrivateData();
	data->ref_finish = reference_finished;
	data->ref_progress = reference_progress;
	data->ref_id = ref_id;


	HANDLE hFile = CreateFileA(file, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	data->hFile = hFile;

	curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, write_callback);
	curl_easy_setopt(handle, CURLOPT_WRITEDATA, hFile);

	curl_easy_setopt(handle, CURLOPT_PROGRESSFUNCTION, progress_callback);
	curl_easy_setopt(handle, CURLOPT_PROGRESSDATA, (char *)data);

	curl_easy_setopt(handle, CURLOPT_NOPROGRESS, 0);
	curl_easy_setopt(handle, CURLOPT_PRIVATE, (char *)data);

	curl_easy_setopt(handle, CURLOPT_ENCODING, "");

	return handle;
}

void DispatchHandle(CURL *handle)
{
	curl_multi_add_handle(multi_handle, handle);
	++num_curl_handles;
}

// gmodeworld_downloadfile(localzipfile, remotezipfile, completefunc, progressfunc, reference, postdata {})
LUA_FUNCTION(gmodeworld_downloadfile)
{
	ILuaInterface *gLua = Lua();

	gLua->CheckType(6, GLua::TYPE_TABLE);

	CURL *handle = GetNewHandle(gLua);

	struct curl_httppost *formpost = NULL;
	struct curl_httppost *lastptr = NULL;

	ILuaObject *table = gLua->GetObject(6);

	CUtlLuaVector *tablemembers = table->GetMembers();

	for (int i = 0; i < tablemembers->Size(); i++ )
	{
		LuaKeyValue kv = (*tablemembers)[i];

		curl_formadd(&formpost,
			&lastptr,
			CURLFORM_COPYNAME, kv.pKey->GetString(),
			CURLFORM_COPYCONTENTS, kv.pValue->GetString(),
			CURLFORM_END);
	}

	LuaPrivateData *data = NULL;
	curl_easy_getinfo(handle, CURLINFO_PRIVATE, &data);
	data->post = formpost;

	curl_easy_setopt(handle, CURLOPT_HTTPPOST, formpost);

	if(handle)
		DispatchHandle(handle);
	return 0;
}

// bool gmodeworld_filedownloaded(zipfile)
LUA_FUNCTION(gmodeworld_filedownloaded)
{
	ILuaInterface *gLua = Lua();

	const char *file = GetZipFile(gLua);
	if(!file)
		return 0;

	DWORD dwFile = GetFileAttributesA(file);

	gLua->Push(dwFile != INVALID_FILE_ATTRIBUTES);
	return 1;
}

// string gmodeworld_fileinfo(zipfile)
LUA_FUNCTION(gmodeworld_fileinfo)
{
	ILuaInterface *gLua = Lua();

	ZZIP_DIR *zipdir = GetZipDir(gLua);

	if(!zipdir)
		return 0;

	ZZIP_DIRENT dirent;
	const char *ffile = GetFirstZipDir(zipdir, &dirent);
	if(ffile)
	{
		char infofile[MAX_PATH];
		_snprintf(infofile, sizeof(infofile), "%sinfo.txt", ffile);

		ZZIP_FILE *zfile = zzip_file_open(zipdir, infofile, 0);

		if(zfile)
		{
			ZZIP_STAT zstat;
			zzip_file_stat(zfile, &zstat);

			char *buff = new char[zstat.st_size];
			char *buffptr = buff;
			zzip_size_t bufflen = zstat.st_size;
			zzip_size_t len = 0;

			while( ( len = zzip_file_read(zfile, buffptr, bufflen) ) > 0)
			{
				bufflen -= len;
				buffptr += len;
			}

			zzip_file_close(zfile);
			zzip_closedir(zipdir);

			gLua->Push(buff);

			delete buff;
			return 1;
		}
	}

	zzip_closedir(zipdir);

	gLua->Push(false);
	return 1;
}

// gmodeworld_setaddon(zipfile, true/false)
LUA_FUNCTION(gmodeworld_setaddon)
{
	ILuaInterface *gLua = Lua();

	gLua->CheckType(2, GLua::TYPE_BOOL);

	ZZIP_DIR *zipdir = GetZipDir(gLua);

	if(!zipdir)
		return 0;

	bool enable = gLua->GetBool(2);
	bool result = false;

	ZZIP_DIRENT dirent;
	const char *ffile = GetFirstZipDir(zipdir, &dirent);
	if(ffile)
	{
		char infofile[MAX_PATH];
		char newinfofile[MAX_PATH];
		_snprintf(infofile, sizeof(infofile), "%saddons\\%sinfo.txt", modulemanager->GetBaseFolder(), ffile);
		_snprintf(newinfofile, sizeof(infofile), "%saddons\\%s_info.txt", modulemanager->GetBaseFolder(), ffile);

		if(enable)
			result = MoveFileA(newinfofile, infofile);
		else
			result = MoveFileA(infofile, newinfofile);
	}

	zzip_closedir(zipdir);

	gLua->Push(result);
	return 1;
}

// bool, bool gmodeworld_fileinstalled(zipfile)
// first bool is if the zip is installed
// second bool is if the addon is installed
LUA_FUNCTION(gmodeworld_fileinstalled)
{
	ILuaInterface *gLua = Lua();

	ZZIP_DIR *zipdir = GetZipDir(gLua);

	if(!zipdir)
		return 0;

	ZZIP_DIRENT dirent;
	const char *ffile = GetFirstZipDir(zipdir, &dirent);
	if(ffile)
	{
		char basefile[MAX_PATH];
		_snprintf(basefile, sizeof(basefile), "%saddons\\%s", modulemanager->GetBaseFolder(), ffile);

		DWORD dwFile = GetFileAttributesA(basefile);
		gLua->Push(dwFile != INVALID_FILE_ATTRIBUTES);

		_snprintf(basefile, sizeof(basefile), "%saddons\\%sinfo.txt", modulemanager->GetBaseFolder(), ffile);

		dwFile = GetFileAttributesA(basefile);
		gLua->Push(dwFile != INVALID_FILE_ATTRIBUTES);
	} else {
		gLua->Push(false);
		gLua->Push(false);
	}


	zzip_closedir(zipdir);


	return 2;
}

struct ExtractPrivateData
{
	bool complete;
	char *file;
	double size;
	double written;
	double ackwritten;

	int ref_complete;
	int ref_progress;
	int ref_id;
};
std::list<ExtractPrivateData *> extracting;

DWORD WINAPI ThreadedExtract( LPVOID lpParam ) 
{
	ExtractPrivateData *data = (ExtractPrivateData *) lpParam;

	ZZIP_DIR *zipdir = zzip_opendir(data->file);
	if(!zipdir)
	{
		data->complete = true;
		return 0;
	}

	ZZIP_DIRENT dirent;
	while(zzip_dir_read(zipdir, &dirent))
	{
		char *name = dirent.d_name;
		size_t len = strlen(name);

		char gmodfile[MAX_PATH];
		_snprintf(gmodfile, sizeof(gmodfile), "%saddons\\%s", modulemanager->GetBaseFolder(), name);

		if(name[len-1] == '/')
		{
			CreateDirectoryA(gmodfile, NULL);
		} else {
			ZZIP_FILE *file = zzip_file_open(zipdir, name, 0);

			if(!file)
				break;

			HANDLE hFile = CreateFileA(gmodfile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

			char buff[8192];
			zzip_size_t len = 0;

			while( ( len = zzip_file_read(file, buff, sizeof(buff)) ) > 0)
			{
				DWORD dwWritten;
				data->written += len;
				WriteFile(hFile, buff, len, &dwWritten, NULL);
			}

			zzip_file_close(file);
			CloseHandle(hFile);
		}
	}

	data->complete = true;
	zzip_closedir(zipdir);
	free(data->file);

	return 0;
}

// gmodeworld_extractzip(zipfile, completefunc, progressfunc, reference, 
LUA_FUNCTION(gmodeworld_extractzip)
{
	ILuaInterface *gLua = Lua();

	gLua->CheckType(2, GLua::TYPE_FUNCTION);
	gLua->CheckType(3, GLua::TYPE_FUNCTION);

	const char *file = GetZipFile(gLua);
	if(!file)
		return 0;

	ZZIP_DIR *zipdir = zzip_opendir(file);
	if(!zipdir)
		return 0;

	DWORD uncompressedsize = 0;

	ZZIP_DIRENT dirent;
	while(zzip_dir_read(zipdir, &dirent))
	{
		ZZIP_STAT zstat;
		zzip_dir_stat(zipdir, dirent.d_name, &zstat, 0);

		uncompressedsize += zstat.st_size;
	}

	zzip_closedir(zipdir);

	ExtractPrivateData *data = new ExtractPrivateData();
	data->complete = false;
	data->size = uncompressedsize;
	data->written = 0;
	data->ackwritten = 0;

	data->ref_complete = gLua->GetReference(2);
	data->ref_progress = gLua->GetReference(3);
	data->ref_id = gLua->GetReference(4);

	data->file = strdup(file);

	CreateThread(NULL, 0, ThreadedExtract, data, 0, NULL);

	extracting.push_front(data);

	gLua->Push(true);
	return 1;
}

LUA_FUNCTION(gmodeworld_handler)
{
	curl_poll(L);

	ILuaInterface *gLua = Lua();

	if(!extracting.empty())
	{
		std::list<ExtractPrivateData *>::iterator iter = extracting.begin();

		while (iter != extracting.end())
		{
			ExtractPrivateData *data = *iter;

			if(!data->complete)
			{
				if(data->written != data->ackwritten)
				{
					gLua->PushReference(data->ref_progress);

					gLua->PushReference(data->ref_id);
					gLua->PushDouble(data->size);
					gLua->PushDouble(data->written);

					gLua->Call(3, 0);

					data->ackwritten = data->written;
				}
				++iter;
			}
			else
			{
				gLua->PushReference(data->ref_complete);

				gLua->PushReference(data->ref_id);
				gLua->PushDouble(data->size);

				gLua->Call(2, 0);

				iter = extracting.erase(iter);

				gLua->FreeReference(data->ref_complete);
				gLua->FreeReference(data->ref_progress);
				gLua->FreeReference(data->ref_id);
				
				delete data;
			}
		}

	}

	return 0;
}

// luashared_printstats()
LUA_FUNCTION(luashared_printstats)
{
	ILuaInterface *gLua = Lua();

	if(!lua_shared)
		return 0;

	lua_shared->PrintLuaStats();

	return 0;
}

// luashared_loadaddons()
LUA_FUNCTION(luashared_loadaddons)
{
	ILuaInterface *gLua = Lua();

	if(!lua_shared)
		return 0;

	AddonVector *addonlist = lua_shared->GetLuaAddons();

	// is it worth clearing the search path and letting it rebuild?
	/*
	for( int i=0; i < addonlist->Count(); i++ )
	{
		char *addon = addonlist->Element(i);
		gLua->Msg("addon: %s\n", addon);
	}*/

	addonlist->RemoveAll();

	lua_shared->LoadAddonInfo();

	return 0;
}

int Start( lua_State* L )
{
	ILuaInterface *gLua = Lua();

	if((int)L != 0x9660000)
	{
		gLua->Msg("Invalid gmod version or tried to load in wrong lua state!\n");
		return 1;
	}

	CreateInterfaceFn luaFactory = Sys_GetFactory( "lua_shared.dll" );
	lua_shared = (ILuaShared *)luaFactory(GARRYSSECRETLUAINTERFACE, NULL);

	CreateInterfaceFn fsFactory = Sys_GetFactory( "filesystem_steam.dll" );
	if(fsFactory)
	{
		filesystem = (IFileSystem *)fsFactory(FILESYSTEM_INTERFACE_VERSION, NULL);
	} else {
		gLua->Msg("No steam filesystem!\n");
		//return 1;
	}
	g_pLua = gLua;

	curl_global_init(CURL_GLOBAL_ALL);

	multi_handle = curl_multi_init();

	gLua->SetGlobal("gmodeworld_downloadfile", gmodeworld_downloadfile);
	gLua->SetGlobal("gmodeworld_setaddon", gmodeworld_setaddon);
	gLua->SetGlobal("gmodeworld_filedownloaded", gmodeworld_filedownloaded);
	gLua->SetGlobal("gmodeworld_fileinstalled", gmodeworld_fileinstalled);
	gLua->SetGlobal("gmodeworld_fileinfo", gmodeworld_fileinfo);
	gLua->SetGlobal("gmodeworld_extractzip", gmodeworld_extractzip);

	gLua->SetGlobal("luashared_printstats", luashared_printstats);
	gLua->SetGlobal("luashared_loadaddons", luashared_loadaddons);

	ILuaObject *hookt = gLua->GetGlobal("hook");
	ILuaObject *addf = hookt->GetMember("Add");
	addf->Push();
	gLua->Push("Think");
	gLua->Push("GmodeWorldHandler");
	gLua->Push(gmodeworld_handler);
	gLua->Call(3);

	gLua->Msg("GmodeWorld loaded. libcurl version: %s\n", curl_version());
	return 0;
}

int Close( lua_State* L )
{
	ILuaInterface *gLua = Lua();

	if(num_curl_handles > 0 || !extracting.empty())
	{
		gLua->Msg("Blocking close until GmodeWorld is finished!");
	}

	while(num_curl_handles > 0 || !extracting.empty())
	{
		gmodeworld_handler(L);
		Sleep(100);
	}

	curl_multi_cleanup(multi_handle);

	curl_global_cleanup();
	return 0;
}